local ffi = require("ffi")
local bit = require("bit")

--require "autobatch"
--love.graphics.setDefaultFilter("nearest")

local map -- stores tiledata
local mapWidth, mapHeight -- width and height in tiles
 
local mapX, mapY -- view x,y in tiles. can be a fractional value like 3.25.
 
local tilesDisplayWidth, tilesDisplayHeight -- number of tiles to show
local zoomX, zoomY
 
local tilesetImage
local tileSize -- size of tiles in pixels
local tileQuads = {} -- parts of the tileset used for different tiles
local tilesetSprite
local tileSize = 32

local FPS = 60
local memstart = 0
local memcurrent = 0
local mempeak = 0
local memarray = 0


local mapWidth = 300
local mapHeight = 300
local tilesDisplayWidth = 34
local tilesDisplayHeight = 20
local buffersize = 1 * (mapHeight + 1)

local lifechance = 5
local currentmap = 1
local totalcells = 0

local lifeskipupdates = 5
local lifecurrentupdate = 0

  local td = 0
  local ty = 0
  local ts = 0

  local function resetty()
    td = 15 -- text delta for Y
    ty = 20 -- start pos
    ts = -1  -- step
  end

  local function getty()
    ts = ts + 1
    return (ty + td * ts)
  end

function GetMemoryUsage()
    return collectgarbage("count")
end

function GetMemoryString(v)
--    return tostring(math.ceil(v/1024))..'M'
    return tostring(math.ceil(v))..'K'
end

function UpdateMemoryUsage()
    memcurrent = GetMemoryUsage()
    if memcurrent > mempeak then mempeak = memcurrent end
end

function GetMemoryStatus()
--    return 'current = '..GetMemoryString(memcurrent)..', start = '..GetMemoryString(memstart)..'->'..GetMemoryString(memarray)..'->'..GetMemoryString(membuffer)..', peak = '..GetMemoryString(mempeak)
    return 'current = '..GetMemoryString(memcurrent)..', start = '..GetMemoryString(memstart)..'->'..GetMemoryString(memarray)..', peak = '..GetMemoryString(mempeak)
end

function love.load()

  setupMap()
  setupMapView()
  setupTileset()
--  love.graphics.setFont(12)
  min_dt = 1/FPS
  next_time = love.timer.getTime()
end

function setupMap()
    memstart = GetMemoryUsage()
    map = {}
    for i = 1,2 do
	map[i] = {}
        for x = 1,mapWidth do
	    map[i][x] = ffi.new("char[?]", buffersize)
	    for y = 1,mapHeight do
		if i == 1 then
    		    if love.math.random(0, lifechance) == lifechance then map[i][x][y] = 1 else map[i][x][y] = 0 end
		else
		    map[i][x][y] = 0
		end
	    end
	end
    end
    memarray = GetMemoryUsage()
end

function UpdateMap()
    local src = 1
    local dst = 2

    local n = {}

    local cellsaround = 0

    for i = 1,3 do
	n[i] = {}
	for j = 1,3 do
	    n[i][j] = {}
	end
    end
    if currentmap == 2 then
	src = 2
	dst = 1
	currentmap = 1
    else
	currentmap = 2
    end

    totalcells = 0
    for x = 1,mapWidth do
	for y = 1,mapHeight do
--	    if map[src][x][y] == 2 then map[src][x][y] = 0 end
	    for i = 1,3 do
		for j = 1,3 do
		    n[i][j][1] = x - i + 2
		    n[i][j][2] = y - j + 2
		    if n[i][j][1] < 1 then n[i][j][1] = mapWidth end
		    if n[i][j][2] < 1 then n[i][j][2] = mapHeight end
		    if n[i][j][1] > mapWidth then n[i][j][1] = 1 end
		    if n[i][j][2] > mapHeight then n[i][j][2] = 1 end
		end
	    end
	    cellsaround = 0
	    for i = 1,3 do
		for j = 1,3 do
		    if (i == 2) and (j == 2) then
			local k = 0
		    else
			if map[src][ n[i][j][1] ][ n[i][j][2] ] == 1 then cellsaround = cellsaround + 1 end
		    end
		end
	    end
	    map[dst][x][y] = map[src][x][y]
	    if (map[src][x][y] == 0) and (cellsaround == 3) then 
		map[dst][x][y] = 1 
	    end
	    if (cellsaround < 2) or (cellsaround > 3) then 
		map[dst][x][y] = 0
	    end
	    if map[dst][x][y] == 1 then totalcells = totalcells + 1 end
	end
  end
end
 
function setupMapView()
  mapX = 1
  mapY = 1
 
  zoomX = 1
  zoomY = 1
end
 
function setupTileset()
  tilesetImage = love.graphics.newImage( "tile.png" )
--  tilesetImage = love.graphics.newImage( "tileset.png" )
  tilesetImage:setFilter("nearest", "linear") -- this "linear filter" removes some artifacts if we were to scale the tiles
 
  -- grass
  tileQuads[0] = love.graphics.newQuad(1 * tileSize, 3 * tileSize, tileSize, tileSize,
    tilesetImage:getWidth(), tilesetImage:getHeight())
  -- kitchen floor tile
  tileQuads[1] = love.graphics.newQuad(0 * tileSize, 3 * tileSize, tileSize, tileSize,
    tilesetImage:getWidth(), tilesetImage:getHeight())
  -- parquet flooring
 
  tilesetBatch = love.graphics.newSpriteBatch(tilesetImage, tilesDisplayWidth * tilesDisplayHeight)
 
  updateTilesetBatch()
end
 
function updateTilesetBatch()
  tilesetBatch:clear()
  for x=0, tilesDisplayWidth-1 do
    for y=0, tilesDisplayHeight-1 do
	local m = map[currentmap][x+math.floor(mapX)][y+math.floor(mapY)]
        tilesetBatch:add(tileQuads[m],x*tileSize, y*tileSize)
    end
  end
  tilesetBatch:flush()
end
 
-- central function for moving the map
function moveMap(dx, dy)
  oldMapX = mapX
  oldMapY = mapY
  mapX = math.max(math.min(mapX + dx, mapWidth - tilesDisplayWidth), 1)
  mapY = math.max(math.min(mapY + dy, mapHeight - tilesDisplayHeight), 1)
  -- only update if we actually moved
  if math.floor(mapX) ~= math.floor(oldMapX) or math.floor(mapY) ~= math.floor(oldMapY) then
    updateTilesetBatch()
  end
end

function love.keypressed(key)
   if key == "escape" then
      love.event.quit()
   end

   if key == "space" then
	collectgarbage("collect")
   end
end
 
function love.update(dt)
  local step = 2

    lifecurrentupdate = lifecurrentupdate + 1
    if lifecurrentupdate > lifeskipupdates then
	lifecurrentupdate = 0
	UpdateMap()
	updateTilesetBatch()
    end

  next_time = next_time + min_dt
  
  if love.keyboard.isDown("up")  then
    moveMap(0, -step * tileSize * dt)
  end
  if love.keyboard.isDown("down")  then
    moveMap(0, step * tileSize * dt)
  end
  if love.keyboard.isDown("left")  then
    moveMap(-step * tileSize * dt, 0)
  end
  if love.keyboard.isDown("right")  then
    moveMap(step * tileSize * dt, 0)
  end
end


 
function love.draw()

  love.graphics.draw(tilesetBatch,
    math.floor(-zoomX*(mapX%1)*tileSize), math.floor(-zoomY*(mapY%1)*tileSize),
    0, zoomX, zoomY)

  UpdateMemoryUsage()

  resetty()
  love.graphics.print('FPS: '..love.timer.getFPS(), 10, getty())
  love.graphics.print('mapX: '..tostring(math.ceil(mapX))..'/'..tostring(mapWidth), 10, getty())
  love.graphics.print('mapY: '..tostring(math.ceil(mapY))..'/'..tostring(mapHeight), 10, getty())
  love.graphics.print('Mem: '..GetMemoryStatus() , 10, getty())
  love.graphics.print('Cells : '..tostring(math.ceil(totalcells)) , 10, getty())

  local cur_time = love.timer.getTime()
   if next_time <= cur_time then
      next_time = cur_time
      return
   end
   love.timer.sleep(next_time - cur_time)
end
